"""Verilog testbench code generation."""
import uuid


def _uniquify(s, l):
    while s in l:
        s += f"_{uuid.uuid4()}"
    return s


def generate_testbench(output_dir, name, inputs, outputs, simulator):
    """
    Generate testbench code that can be used to simulate a circuit.

    Parameters
    ----------
    output_dir: pathlib.Path
            Path to save tb and any extraneous files to.
    name: str
            The name of the module to be simulated.
    inputs: list of str
            The inputs to the module.
    outputs: list of str
            The outputs to the module.
    simulator: str
            The simulator tool to generate a testbench for.

    Returns
    -------
    str
            The generated testbench code.

    """
    if simulator == "verilator":
        tick = _uniquify("tick", inputs + outputs)
        tb = f"module {name}_tb({tick});\n\n"
        tb += f"  input {tick};\n"
        first_sim = _uniquify("first_sim", inputs + outputs)
        tb += f"  reg {first_sim};\n\n"
    else:
        tb = "module tb;\n\n"
    tb += "\n".join(f"  wire {i};" for i in inputs) + "\n\n"
    tb += "\n".join(f"  wire {o};" for o in outputs) + "\n\n"

    tb += f"  {name} {name}_inst(\n"
    tb += ",\n".join(f"    .{i}({i})" for i in inputs + outputs)
    tb += "\n  );\n\n"

    ret = _uniquify("ret", inputs + outputs)
    tb += f"  integer {ret};\n\n"

    input_vector = _uniquify("input_vector", inputs + outputs)
    tb += f"  reg [{len(inputs)-1}:0] {input_vector};\n\n"

    input_concat = "{" + ", ".join(inputs) + "}"
    output_concat = "{" + ", ".join(outputs) + "}"

    tb += f"  assign {input_concat} = {input_vector};\n\n"

    infile = _uniquify("infile", inputs + outputs)
    outfile = _uniquify("outfile", inputs + outputs)
    infile_pointer = _uniquify("infile_pointer", inputs + outputs)
    outfile_pointer = _uniquify("outfile_pointer", inputs + outputs)

    tb += f"  reg [999:0] {infile};\n"
    tb += f"  reg [999:0] {outfile};\n\n"
    tb += f"  integer {infile_pointer};\n"
    tb += f"  integer {outfile_pointer};\n\n"

    tb += "  initial begin\n"
    tb += f'    if (!$value$plusargs("input_file=%s", {infile})) begin\n'
    tb += '      $display("Error opening input file");\n'
    tb += "      $finish;\n"
    tb += "  end\n"
    tb += f'    if (!$value$plusargs("output_file=%s", {outfile})) begin\n'
    tb += '      $display("Error opening output file");\n'
    tb += "      $finish;\n"
    tb += "  end\n"
    tb += f'    {infile_pointer} = $fopen({infile}, "r");\n'
    tb += f'    {outfile_pointer} = $fopen({outfile}, "w");\n'

    if simulator == "verilator":
        tb += f"    {first_sim} = 1;\n"
        tb += "  end\n\n"
        tb += f"  always@(posedge {tick}) begin\n"
        tb += f"    if ({first_sim}) begin\n"
        tb += f"      {first_sim} <= 0;\n"
        tb += "    end\n"
        tb += "    else begin\n"
        tb += f'      $fdisplay({outfile_pointer}, "%b", {output_concat});\n'
        tb += "    end\n"
        tb += f"    if ($feof({infile_pointer})) begin\n"
        tb += f"      $fclose({infile_pointer});\n"
        tb += f"      $fclose({outfile_pointer});\n"
        tb += "      $finish;\n"
        tb += "    end\n"
        tb += f'    $fscanf({infile_pointer}, "%b\\n", {input_vector});\n'
        tb += "  end\n"

        c = '#include "Vtb.h"\n'
        c += '#include "verilated.h"\n'
        c += '#include "verilated_vcd_c.h"\n\n'
        c += "int main(int argc, char **argv, char **env) {\n"
        c += "  Verilated::commandArgs(argc, argv);\n"
        c += "  Vtb* top = new Vtb;\n"
        c += "  while(!Verilated::gotFinish()) {\n"
        c += f"    top->{tick} = 1;\n"
        c += "    top->eval();\n"
        c += f"    top->{tick} = 0;\n"
        c += "    top->eval();\n"
        c += "  }\n"
        c += "}\n"
        with open(output_dir / "tb.cpp", "w") as f:
            f.write(c)
    else:
        tb += f'    {infile_pointer} = $fopen({infile}, "r");\n'
        tb += f'    {outfile_pointer} = $fopen({outfile}, "w");\n'
        tb += f"    while (!$feof({infile_pointer})) begin\n"
        tb += f'      ret = $fscanf({infile_pointer}, "%b\\n", {input_vector});\n'
        tb += f'      #1 $fdisplay({outfile_pointer}, "%b", {output_concat});\n'
        tb += "    end\n"
        tb += f"    $fclose({infile_pointer});\n"
        tb += f"    $fclose({outfile_pointer});\n"
        tb += "  end\n"
    tb += "endmodule\n"

    with open(output_dir / "tb.v", "w") as f:
        f.write(tb)



"""
Perform gate-level simulations from python.

Examples
--------
Create a circuit to simulate.

>>> import circuitgraph as cg
>>> c = cg.from_lib("c17")

Initialize a simulator object for this circuit.

>>> from circuitsim import CircuitSimulator
>>> simulator = CircuitSimulator(c)

Simulate some vectors

>>> vectors = [
...     {"N1": 0, "N2": 1, "N3": 0, "N6": 1, "N7": 0},
...     {"N1": 1, "N2": 0, "N3": 0, "N6": 1, "N7": 1}
... ]
>>> result = simulator.simulate(vectors)
>>> result[0] == {"N22": True, "N23": True}
True
>>> result[1] == {"N22": False, "N23": True}
True

Perform x-based simulation. Note that strings must be used instead of
`int`/`bool`s.

>>> vectors = [{"N1": "0", "N2": "0", "N3": "x", "N6": "x", "N7": "1"}]
>>> result = simulator.simulate(vectors, allow_x=True)
>>> result[0] == {"N22": "0", "N23": "x"}
True

"""

from circuitsim.simulation import CircuitSimulator
from circuitsim.simulators import SimulationCompilationError
from circuitsim.simulators import SimulationExecutionError



"""Interface with simulators."""
import math
import subprocess
from multiprocessing import Pool
from pathlib import Path

available_simulators = ["iverilog", "verilator", "vcs"]

compile_args = {
    "vcs": ["vcs", "-full64"],
    "iverilog": ["iverilog"],
    "verilator": ["verilator", "--exe", "tb.cpp", "--cc"],
}
post_compile_args = {"verilator": ["make", "-C", "obj_dir", "-f", "Vtb.mk"]}
simulate_args = {
    "vcs": ["./simv", "-q"],
    "iverilog": ["vvp", "a.out"],
    "verilator": ["./obj_dir/Vtb"],
}


class SimulationCompilationError(Exception):
    """Thrown when a simulation fails during compilation."""


class SimulationExecutionError(Exception):
    """Thrown when a simulation fails during execution."""


def _convert_value(v, allow_x=False):
    if v in [True, False, 0, 1]:
        return str(int(v))
    if isinstance(v, str) and v in "01":
        return v
    if allow_x and v.lower() == "x":
        return v.lower()
    raise ValueError(f"Unknown input value '{v}'")


def _convert_values(values, allow_x=False):
    return {k: _convert_value(v, allow_x=allow_x) for k, v in values.items()}


def compile_simulator(simulator, netlists, working_dir):
    """
    Compile a circuit simulation.

    Parameters
    ----------
    simulator: str
            The simulator to use.
    netlist: list of str or pathlib.Path
            The netlists to compile.
    working_dir: str or pathlib.Path
            The path to comiple the simulation in.

    """
    netlists = [str(Path(n).absolute()) for n in netlists]
    working_dir = Path(working_dir)
    with open(working_dir / "compile.log", "w+") as f:
        try:
            subprocess.run(
                compile_args[simulator] + netlists,
                cwd=working_dir,
                stdout=f,
                stderr=f,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            f.seek(0)
            message = f.read()
            raise SimulationCompilationError(message) from e

    if simulator in post_compile_args:
        with open(working_dir / "post_compile.log", "w+") as f:
            try:
                subprocess.run(
                    post_compile_args[simulator],
                    cwd=working_dir,
                    stdout=f,
                    stderr=f,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                f.seek(0)
                message = f.read()
                raise SimulationCompilationError(message) from e


def _run_process(p, simulator, working_dir):
    with open(working_dir / f"simulate_{p}.log", "w") as f:
        return subprocess.run(
            simulate_args[simulator]
            + [f"+input_file=input_file_{p}.txt", f"+output_file=output_file_{p}.txt"],
            cwd=working_dir,
            stdout=f,
            stderr=f,
            check=True,
        )


def execute_simulator(
    simulator, inputs, vectors, working_dir, num_processes, allow_x=False
):
    """
    Execute a compiled simulation.

    Parameters
    ----------
    simulator; str
            The simulator to use.
    inputs: list of str
            The inputs to the module.
    vectors: list of dict of str:bool
            The vectors to simulate. Each vector is represented as a
            dictionary mapping an input to a logical value.
    working_dir: str or pathlib.Path
            The directory to compile and run the simulation in.
    num_process: int
            The number of simulation processes to run. Specifying a
            number more than 1 will cause multiple simulations to be
            executed in parallel.
    allow_x: bool
            If True, the inputs/outputs can contain "don't care" or "x"
            values in addition to 0 and 1. Specify an don't care value
            for an input by setting it to "x".

    """
    working_dir = Path(working_dir)

    vectors_per_process = math.ceil(len(vectors) / num_processes)
    for p, v in enumerate(
        [
            vectors[i : i + vectors_per_process]
            for i in range(0, len(vectors), vectors_per_process)
        ]
    ):
        with open(working_dir / f"input_file_{p}.txt", "w") as f:
            for vector in v:
                vector = _convert_values(vector, allow_x=allow_x)
                try:
                    f.write("".join(vector[i] for i in inputs) + "\n")
                except KeyError as e:
                    missing_inputs = ", ".join(set(inputs) - set(vector))
                    if len(missing_inputs) > 20:
                        missing_inputs = missing_inputs[:17] + "..."
                    raise ValueError(
                        "Value not provided for inputs '{missing_inputs}'"
                    ) from e

    if num_processes > 1:
        with Pool(num_processes) as pool:
            results = pool.starmap_async(
                _run_process,
                [(p, simulator, working_dir) for p in range(num_processes)],
            )

            sim_results = results.get()
            for sim_res in sim_results:
                if sim_res.returncode != 0:
                    with open(working_dir / "simulate.log") as f:
                        message = f.read()
                    raise SimulationExecutionError(message)
    else:
        sim_res = _run_process(0, simulator, working_dir)


def parse_simulation_output(working_dir, outputs, num_processes, allow_x=False):
    """
    Parse the output of a simulation.

    Parameters
    ----------
    working_dir: str or pathlib.Path
            The directory that the simulation was run from.

    Returns
    -------
    list of dict of str to bool
            The simulation results, as a list of dictionaries mapping
            outputs to values.
    num_process: int
            The number of simulation processes that were run.
    allow_x: bool
            If True, the outputs can contain "don't care" or "x"
            values in addition to 0 and 1. The outputs will then be
            returned as a dict of str:str, where each output is either
            "0", "1", or "x".

    Returns
    -------
    list o dict o str:bool or str:str
            The simulation outputs. Each vector is represented as a
            dictionary mapping an output to a logical value. If `allow_x`
            is True, then instead of logic values, each output will be
            mapped to either "0", "1", or "x".

    """
    working_dir = Path(working_dir)
    vectors = []
    for p in range(num_processes):
        with open(working_dir / f"output_file_{p}.txt") as f:
            for line in f:
                if allow_x:
                    if not set(line.strip().lower()) <= {"0", "1", "x"}:
                        raise ValueError(f"Unknown value in simulation output: {line}")
                    vectors.append(dict(zip(outputs, line.strip().lower())))
                else:
                    if not set(line.strip()) <= {"0", "1"}:
                        raise ValueError(f"Unknown value in simulation output: {line}")
                    vectors.append(
                        {o: bool(int(v)) for o, v in zip(outputs, line.strip())}
                    )

    return vectors



import setuptools

with open("README.md") as fh:
    long_description = fh.read()

setuptools.setup(
    name="circuitsim",
    version="0.0.1",
    author="Ruben Purdy",
    author_email="rpurdy@andrew.cmu.edu",
    description="Run gate-level HDL simulations from python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/circuitgraph/circuitsim",
    project_urls={
        "Documentation": "https://circuitgraph.github.io/circuitsim/",
        "Source": "https://github.com/circuitgraph/circuitsim",
    },
    packages=["circuitsim"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["circuitgraph>=0.2.0", "natsort"],
)



"""High-level simulation API."""
import tempfile
from pathlib import Path

import circuitgraph as cg
from natsort import natsorted

from circuitsim.codegen import generate_testbench
from circuitsim.simulators import (
    available_simulators,
    compile_simulator,
    execute_simulator,
    parse_simulation_output,
)


class CircuitSimulator:
    """Class for circuit simulation."""

    def __init__(self, ckt, working_dir=None, simulator="iverilog"):
        """
        Create new simulator.

        Parameters
        ----------
        ckt: circuitgraph.Circuit
                The circuit to simulate.
        working_dir: str or pathlib.Path
                The directory to write simulation files to. If `None`, a
                temporary directory will be used.
        simulator: str
                The simulator to use. One of ['iverilog', 'verilator', 'vcs'].

        """
        if simulator not in available_simulators:
            raise ValueError(
                f"Invalid simulator '{simulator}'. Must be one of "
                f"{available_simulators}."
            )
        if ckt.is_cyclic():
            raise ValueError("Cannot simulate cyclic circuit")
        self.simulator = simulator
        self.ckt = ckt
        if working_dir is None:
            self.temp_dir = tempfile.TemporaryDirectory(
                prefix=f"circuitsim_{ckt.name}_"
            )
            self.working_dir = Path(self.temp_dir.name)
        else:
            self.temp_dir = None
            self.working_dir = Path(working_dir)
            self.working_dir.mkdir(exist_ok=True)
        self.inputs = natsorted(list(ckt.inputs()))
        self.outputs = natsorted(list(ckt.outputs()))
        self._initialized = False

    def __del__(self):
        """Remove temporary directory if necessary."""
        if self.temp_dir:
            self.temp_dir.cleanup()

    def _initialize_simulator(self):
        generate_testbench(
            self.working_dir, self.ckt.name, self.inputs, self.outputs, self.simulator
        )
        cg.to_file(self.ckt, self.working_dir / f"{self.ckt.name}.v")
        netlists = [self.working_dir / "tb.v", self.working_dir / f"{self.ckt.name}.v"]
        compile_simulator(self.simulator, netlists, self.working_dir)
        self._initialized = True

    def simulate(self, vectors, num_processes=1, allow_x=False):
        """
        Execute the simulator on a list of vectors.

        Parameters
        ----------
        vectors: list of dict of str:bool
                The vectors to simulate. Each vector is represented as a
                dictionary mapping an input to a logical value.
        num_process: int
                The number of simulation processes to run. Specifying a
                number more than 1 will cause multiple simulations to be
                executed in parallel.
        allow_x: bool
                If True, the inputs/outputs can contain "don't care" or "x"
                values in addition to 0 and 1. Specify an don't care value
                for an input by setting it to "x". The outputs will then be
                returned as a dict of str:str, where each output is either
                "0", "1", or "x".

        Returns
        -------
        list of dict of str:bool or str:str
                The simulation outputs. Each vector is represented as a
                dictionary mapping an output to a logical value. If `allow_x`
                is True, then instead of logic values, each output will be
                mapped to either "0", "1", or "x".

        """
        if not self._initialized:
            self._initialize_simulator()
        if self.simulator == "verilator" and allow_x:
            raise ValueError("Cannot use x-based simulation with verilator")
        execute_simulator(
            self.simulator,
            self.inputs,
            vectors,
            self.working_dir,
            num_processes,
            allow_x=allow_x,
        )
        return parse_simulation_output(
            self.working_dir, self.outputs, num_processes, allow_x=allow_x
        )
